use std::ops::Deref;

use anyhow::{anyhow, Error};
use bonsaidb::{
    core::schema::{Collection, SerializedCollection},
    local::Database,
};
use itertools::Itertools;
use serde::{Deserialize, Serialize};
use ubisync_lib::types::{Peer, PeerId};

use crate::state::database::{as_key::AsKey, StateDB};

#[derive(Debug, Serialize, Deserialize, Collection, PartialEq, Clone)]
#[collection(name = "peers", views = [])]
pub(super) struct DbPeer {
    #[natural_id]
    pub(super) id: AsKey<PeerId>,
    pub(super) name: Option<String>,
}

pub(crate) struct Peers<'a> {
    parent: &'a StateDB,
    db: &'a Database,
}


impl<'a> Peers<'a> {
    pub const fn new(parent: &'a StateDB, bonsai: &'a Database) -> Self {
        Self { parent, db: bonsai }
    }

    pub fn add(&self, peer: Peer) -> anyhow::Result<()> {
        if self.get(peer.id())?.is_some() {
            Err(Error::msg("Peer already exists"))
        } else {
            DbPeer::push(
                DbPeer {
                    id: AsKey::new(peer.id()),
                    name: peer.name(),
                },
                self.db,
            )
            .map(|_| ())
            .map_err(|e| anyhow!(e))
        }
    }

    pub fn get(&self, id: PeerId) -> anyhow::Result<Option<Peer>> {
        let fam = match self.parent.families().get_by_peer(id.clone())? {
            Some(fam_id) => self.parent.families().get(fam_id)?,
            None => None,
        };

        DbPeer::get(&AsKey::new(id), self.db)
            .map(|doc| {
                doc.map(|peer| Peer::new((*peer.contents.id).clone(), peer.contents.name, fam))
            })
            .map_err(|e| anyhow!(e))
    }

    pub fn get_all(&self) -> anyhow::Result<Vec<Peer>> {
        DbPeer::all(self.db)
            .query()
            .map(|peers| {
                peers
                    .iter()
                    .map(|p| {
                        let family = match self
                            .parent
                            .families()
                            .get_by_peer(p.contents.id.deref().clone())
                        {
                            Ok(Some(fam_id)) => self.parent.families().get(fam_id).ok().flatten(),
                            _ => None,
                        };

                        Peer::new(
                            p.contents.id.deref().clone(),
                            p.contents.name.clone(),
                            family,
                        )
                    })
                    .collect_vec()
            })
            .map_err(|e| anyhow!(e))
    }

    pub fn set_name(&self, id: PeerId, name: &Option<String>) -> anyhow::Result<()> {
        DbPeer::get(&AsKey::new(id), self.db)
            .map_err(|e| anyhow!(e))?
            .ok_or(Error::msg("Peer not found"))?
            .modify(self.db, |doc| doc.contents.name = name.clone())
            .map_err(|e| anyhow!(e))
    }
}

#[cfg(test)]
mod tests {
    use ubisync_lib::types::{Peer, PeerId};

    use crate::state::database::StateDB;

    #[test]
    fn add_get() {
        let db = StateDB::init(None);

        let peer = Peer::new(PeerId::default(), Some("Peer name".to_string()), None);
        db.peers().add(peer.clone()).unwrap();

        let retrieved_peer = db.peers().get(peer.id()).unwrap();

        assert_eq!(Some(peer), retrieved_peer)
    }

    #[test]
    fn get_all() {
        let db = StateDB::init(None);

        let peer = Peer::new(PeerId::default(), Some("Peer name".to_string()), None);
        db.peers().add(peer.clone()).unwrap();

        let all_peers = db.peers().get_all().unwrap();
        assert_eq!(all_peers, vec![peer]);
    }

    #[test]
    fn set_peer_name() {
        let db = StateDB::init(None);

        let peer = Peer::new(PeerId::default(), Some("Peer name".to_string()), None);
        db.peers().add(peer.clone()).unwrap();
        db.peers()
            .set_name(peer.id().clone(), &Some("New peer name".to_string()))
            .unwrap();

        let retrieved_peer = db.peers().get(peer.id()).unwrap();

        assert_eq!(
            Some(Peer::new(
                peer.id(),
                Some("New peer name".to_string()),
                None
            )),
            retrieved_peer
        )
    }
}
